
/*****************************************************************************
 mdns.c - mDNS related calls
 
 (c) 2015 Copyright HP Development Company, LP

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 of the Software, and to permit persons to whom the Software is furnished to do
 so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 Client/Server generic message format (see messaging-protocol.doc):

 Author: Sanjay Kumar
 \*****************************************************************************/

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <net/if.h>
#include "mdns.h"
#include <unistd.h>
#include <sys/ioctl.h>

/* Convert "www.google.com" to "3www6google3com". */
static int mdns_convert_name_to_dns(const char *name, int name_size, char *dns_name)
{
    int i, x = 0;
    char *p = dns_name;

    if (name == 0 || name[0] == 0)
        return 0;

    for (i = 0; i < name_size && name[i]; i++)
    {
        if (name[i] == '.')
        {
            *p++ = i - x; /* length */
            for (; x < i; x++)
                *p++ = name[x];
            x++;
        }
    }

    if (i)
    {
        *p++ = i - x; /* length */
        for (; x < i; x++)
            *p++ = name[x];
        x++;
    }

    p[x++] = 0;

    return x; /* return length DOES include null termination */
}

#define MREQN_INIT(_mr, index) do {					\
	(&(_mr))->imr_multiaddr.s_addr = inet_addr("224.0.0.251");	\
	(&(_mr))->imr_address.s_addr = htonl(INADDR_ANY);		\
	(&(_mr))->imr_ifindex = (index);				\
    } while(0)

struct mdns_socket {
    int socket;
    struct if_nameindex *idx;
    struct if_nameindex **good;
};
#define MDNS_SOCKET_INIT { .socket = -1, .idx = NULL, .good = NULL, }

static int get_ipv4_address(const char *iface, struct in_addr *addr)
{
    int s, r;
    struct ifreq ifr;

    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, iface, IFNAMSIZ-1);

    s = socket(AF_INET, SOCK_DGRAM, 0);
    r = ioctl(s, SIOCGIFADDR, &ifr);
    close(s);

    if (r == -1) {
	DBG("error in SIOCGIFADDR for %s: %m\n", iface);
	return MDNS_STATUS_ERROR;
    }

    if (addr != NULL)
	memcpy(addr, &((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr,
	       sizeof(*addr));
    return MDNS_STATUS_OK;
}

static int mdns_open_socket(struct mdns_socket *mdns_sock)
{
    int stat = MDNS_STATUS_ERROR;
    int udp_socket = -1, yes = 1, ifaces;
    char loop = 0, ttl = 255;
    struct sockaddr_in recv_addr , addr;
    struct ip_mreqn mreqn;
    struct if_nameindex *idx;

    DBG("mdns_open_socket entry.\n");

    if ((udp_socket = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
    {
        BUG("unable to create udp socket: %m\n");
        goto bugout;
    }

    /* Get rid of "address already in use" error message. */
    if (setsockopt(udp_socket, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)) == -1)
    {
        BUG("unable to setsockopt: %m\n");
        goto bugout;
    }

    /* Bind the socket to port and IP equal to INADDR_ANY. */
    bzero(&recv_addr, sizeof(recv_addr));
    recv_addr.sin_family = AF_INET;
    recv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    recv_addr.sin_port = htons(5353);
    if (bind(udp_socket, (struct sockaddr *) &recv_addr, sizeof(recv_addr)) == -1)
    {
        BUG("unable to bind udp socket: %m\n");
        goto bugout;
    }

    /* Set multicast loopback off. */
    if (setsockopt(udp_socket, IPPROTO_IP, IP_MULTICAST_LOOP, &loop, sizeof(loop)) == -1)
    {
        BUG("unable to setsockopt: %m\n");
        goto bugout;
    }

    /* Set ttl to 255. Required by mdns. */
    if (setsockopt(udp_socket, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl))== -1)
    {
        BUG("unable to setsockopt: %m\n");
        goto bugout;
    }

    mdns_sock->idx = if_nameindex();
    if (mdns_sock->idx == NULL) {
	BUG("if_nameindex failed: %m\n");
	goto bugout;
    }

    for (idx = mdns_sock->idx, ifaces = 0;
	 idx && (idx->if_index != 0 || idx->if_name != NULL); idx++) {
	ifaces ++;
    }

    mdns_sock->good = calloc(ifaces, sizeof(struct if_nameindex*));
    if (mdns_sock->good == NULL)
	goto bugout;

    for (idx = mdns_sock->idx, ifaces = 0;
	 idx && (idx->if_index != 0 || idx->if_name != NULL); idx++) {

	/* Skip lo and interfaces without IPv4 address */
	if (!strcmp(idx->if_name, "lo"))
	    continue;
	if (get_ipv4_address(idx->if_name, NULL) == MDNS_STATUS_ERROR)
	    continue;

	/* Join the multicast group on each local interface */
	MREQN_INIT(mreqn, idx->if_index);
	if (setsockopt(udp_socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreqn,
		       sizeof(struct ip_mreqn)) == -1) {
	    BUG("unable to add to multicast group for %s: %m\n", idx->if_name);
	} else {
	    mdns_sock->good[ifaces++] = idx;
	    DBG("added multicast group on interface %s\n", idx->if_name);
	}
    }

    if (ifaces == 0) {
	BUG("no interfaces for multicast found\n");
        goto bugout;
    }

    mdns_sock->socket = udp_socket;
    DBG("Socket = [%d]: %m\n", mdns_sock->socket);
    stat = MDNS_STATUS_OK;

bugout:
    if (stat != MDNS_STATUS_OK)
	mdns_close_socket(mdns_sock);
    return stat;
}

static void mdns_close_socket(struct mdns_socket *mdns_sock)
{
    int i;
    struct if_nameindex *idx;
    struct ip_mreqn mreqn;
    if (mdns_sock->socket != -1) {
	for (i = 0; mdns_sock->good[i]; i++) {
	    idx = mdns_sock->good[i];
	    MREQN_INIT(mreqn, idx->if_index);
	    if (setsockopt(mdns_sock->socket, IPPROTO_IP, IP_DROP_MEMBERSHIP,
			   &mreqn, sizeof(struct ip_mreqn)) == -1) {
		BUG("unable to drop multicast group for %s: %m\n", idx->if_name);
	    };
	}
	close(mdns_sock->socket);
	mdns_sock->socket = -1;
    }

    if (mdns_sock->idx != NULL) {
	if_freenameindex(mdns_sock->idx);
	mdns_sock->idx = NULL;
    }

    if (mdns_sock->good != NULL) {
	free(mdns_sock->good);
	mdns_sock->good = NULL;
    }
}

static void mdns_create_query_packet(const char* fqdn, int query_type, char* querybuf, int *length)
{
    int n = 0;
    char header[] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
                                                             // ID/FLAGS/QDCNT/ANCNT/NSCNT/ARCNT

    DBG("mdns_create_query_packet.\n");
    memcpy(querybuf, header, sizeof(header));
    n = sizeof(header);

    n += mdns_convert_name_to_dns(fqdn, strlen(fqdn), querybuf + n);
    querybuf[n++] = 0x00;
    querybuf[n++] = query_type;
    querybuf[n++] = 0x00;
    querybuf[n++] = QCLASS_IN;

    //DBG_DUMP(dnsquery, n);
    *length = n;
}

static int mdns_send_query(struct mdns_socket *mdns_sock, const char *fqdn, int query_type)
{
    char querybuf[256] = {0,};
    int length = 0, i, success;
    int stat = MDNS_STATUS_OK;
    struct sockaddr_in send_addr;

    DBG("mdns_send_query entry.  send socket=%d len=%d\n", mdns_sock->socket, length);

    mdns_create_query_packet(fqdn, query_type, querybuf, &length);

    bzero(&send_addr, sizeof(send_addr));
    send_addr.sin_family = AF_INET;
    send_addr.sin_addr.s_addr = inet_addr("224.0.0.251");
    send_addr.sin_port = htons(5353);

    for (i = 0, success = 0; mdns_sock->good[i]; i++) {
	struct if_nameindex *idx;
	struct ip_mreqn mreqn;
	idx = mdns_sock->good[i];
	MREQN_INIT(mreqn, idx->if_index);
	if (setsockopt(mdns_sock->socket, IPPROTO_IP, IP_MULTICAST_IF, &mreqn, sizeof(mreqn))
	    == -1) {
	    DBG("failed to set IP_MULTICAST_IF to %s\n", idx->if_name);
	    continue;
	}
	if (sendto(mdns_sock->socket, querybuf, length, 0,
		   (struct sockaddr *) &send_addr, sizeof(send_addr))< 0) {
	    DBG("failed to send on %s\n", idx->if_name);
	} else
	    success++;
    }

    if (success == 0)
	stat = MDNS_STATUS_ERROR;

    DBG("mdns_send_query returning with status(%d)...\n", stat);
    return stat;
}

static int mdns_readName(unsigned char* start, unsigned char *Response, char *buf)
{
    int size = 0;
    char *name = buf;
    unsigned char *p = Response;

    while (size = *p++)
    {
        if (size >= 0xC0)
        {
            //Compressed Size. Just ignore it.
            p++; //skip Offset byte
            return (p - Response);
        }
        memcpy(name, p, size);
        name[size] = '.';
        p += size;
        name += size + 1;
    }

    *(name - 1) = '\0';

    DBG("Name = [%s]\n", buf);
    return (p - Response);
}


static unsigned char* mdns_readMDL(unsigned char *p, unsigned char *normalized_mdl, int len)
{
    int i = 0;
    int j = 0;
    int size = 0;

    unsigned char* mdl = normalized_mdl;
    while (i < len)
    {
        size = *p++;
        i += size + 1;

        if (strncmp(p, "mdl=", 4) == 0)
        {
            for (j = 4; j < size; j++)
            {
                if (*(p + j) == ' ')
                    *mdl++ = '_'; //Replace white space with underscore
                else
                    *mdl++ = tolower(*(p + j));
            }

            *mdl++ = '\0';
            break;
        }
        p += size;

    }
    DBG("MDL = [%s]\n", normalized_mdl);
    return p + 4;
}

static void mdns_read_header(char *Response, DNS_PKT_HEADER *h)
{
    h->id          =   Response[0] << 8 | Response[1];
    h->flags       =   Response[2] << 8 | Response[3];
    h->questions   =   Response[4] << 8 | Response[5];
    h->answers     =   Response[6] << 8 | Response[7];
    h->authorities =   Response[8] << 8 | Response[9];
    h->additionals =   Response[10]<< 8 | Response[11];

    DBG("ID=%x flags=%x Q=%x A=%x AUTH=%x ADD=%x\n", h->id, h->flags, h->questions,
            h->answers, h->authorities, h->additionals);

}

static int mdns_parse_respponse(unsigned char *Response, DNS_RECORD *rr)
{
    unsigned char *p = Response;
    unsigned short type = 0, data_len = 0;
    DNS_PKT_HEADER h;
    int i = 0;

    DBG("mdns_parse_respponse entry.\n");
    mdns_read_header(Response, &h);
    p += MDNS_HEADER_SIZE;

    if (h.answers + h.additionals <= 0) {
	DBG("mdns_parse_respponse: no answers");
	return MDNS_STATUS_ERROR;
    }

    for (i = 0; i < h.questions; i++)
    {
        p += mdns_readName(Response, p, rr->name);
        p += 4; //Skip TYPE(2 bytes)/CLASS(2 bytes)
    }

    for (i = 0; i < (h.answers + h.additionals); i++)
    {
        p += mdns_readName(Response, p, rr->name);
        type = (*p << 8  | *(p+1));
        p += 8;  //Skip type(2 bytes)/class(2 bytes)/TTL(4 bytes)

        data_len = ( *p << 8  | *(p+1));
        p += 2;  //Skip data_len(2 bytes)

        switch (type)
        {
            case QTYPE_A:
                sprintf(rr->ip, "%d.%d.%d.%d", p[0], p[1], p[2], p[3]);
                break;
            case QTYPE_TXT:
                mdns_readMDL(p, rr->mdl, data_len);
                break;
            default:
                break;
        }

        p += data_len;
        //DBG("TYPE = %d, Length = %d\n",type, data_len);
    }

    DBG("mdns_parse_respponse returning MDL = %s, IP = %s\n",rr->mdl, rr->ip);
    return MDNS_STATUS_OK;
}

static int mdns_read_single_response(int udp_socket, char *recvbuffer, int recvbufsize)
{
    struct timeval tmo;
    struct sockaddr_in addr;
    socklen_t addrlen = sizeof(addr);
    fd_set master, readfd;
    int len = 0, maxfd = 0, ret = 0;

    DBG("mdns_read_single_response.\n");
    FD_ZERO(&master);
    FD_SET(udp_socket, &master);
    maxfd = udp_socket;
    tmo.tv_sec = 0;
    tmo.tv_usec = 10000;

    readfd = master;
    ret = select(maxfd + 1, &readfd, NULL, NULL, &tmo);
    if (ret > 0)
    {
        bzero(&addr, sizeof(addr));
        if ((len = recvfrom(udp_socket, recvbuffer, recvbufsize, 0, (struct sockaddr *) &addr, &addrlen)) < 0)
        {
            BUG("recvfrom error: (%m)\n");
            ret = -1;
        }
    }

    DBG("mdns_read_single_response exiting with ret = %d\n", ret);
    return ret;
}

static DNS_RECORD *mdns_read_responses(int udp_socket, int mode, const char *question)
{
    int retries = 3, ret = 0;
    char recvbuffer[MAX_MDNS_RESPONSE_LEN] =  { 0, };
    DNS_RECORD *rr = NULL, *head = NULL, *temp = NULL;

    DBG("mdns_read_responses.\n");
    while (1 )
    {
        memset(recvbuffer, 0, sizeof(recvbuffer));
        ret = mdns_read_single_response(udp_socket, recvbuffer, sizeof(recvbuffer));
        if (ret <= 0)
        {
            if (ret == 0 && retries--)     //READ TIMEOUT. Retry few more times.
                continue;
            else
                break;
        }
        else
        {
            temp = (DNS_RECORD *)malloc(sizeof(DNS_RECORD));
            if(temp)
            {
                memset(temp, 0, sizeof(DNS_RECORD));
                if (mdns_parse_respponse(recvbuffer, temp) == MDNS_STATUS_OK &&
		    (mode == MODE_READ_ALL || question == NULL ||
		     !strncmp(question, temp->name, sizeof(temp->name)-1))) {
		    if(head == NULL)
			rr = head = temp;
		    else
		    {
			rr->next = temp;
			rr = rr->next;
		    }
		} else {
		    DBG("Parse error or wrong MDNS name");
		    free(temp);
		    continue;
		}

                if(mode == MODE_READ_SINGLE)
                    break;
            }
        }
    } // while(1)

    DBG("mdns_read_responses returning with (%p).\n", head);
    return head;
}

static int mdns_update_uris(DNS_RECORD *rr, char* uris_buf, int buf_size, int *count)
{
    char tempuri[MAX_URI_LEN] = {0};
    int bytes_read = 0;

    DBG("mdns_update_uris.\n");

    *count = 0;
    memset(uris_buf, 0, buf_size);

    while(rr)
    {
        if (rr->mdl[0] && rr->ip[0] /*&& strstr(rr->mdl, "scanjet")*/)
        {
            memset(tempuri, 0, sizeof(tempuri));
            sprintf(tempuri, "hp:/net/%s?ip=%s&queue=false", rr->mdl, rr->ip);

            //Check whether buffer has enough space to add new URI and check for duplicate URIs.
            if(bytes_read + sizeof(tempuri) < buf_size  && !strstr(uris_buf, tempuri))
            {
                bytes_read += sprintf(uris_buf + bytes_read, "%s;", tempuri);
                (*count)++;
                *(uris_buf + bytes_read) = '\0';
            }
        }
        rr = rr->next;
    }

    DBG("mdns_update_uris Count=[%d] bytes=[%d] URIs = %s\n",*count, bytes_read, uris_buf);
    return bytes_read;
}

static void mdns_rr_cleanup(DNS_RECORD *rr)
{
    DNS_RECORD *temp = NULL;

    DBG("mdns_rr_cleanup entry.\n");
    while(rr)
    {
        temp = rr->next;
        free(rr);
        rr = temp;
    }
}

static DNS_RECORD* send_and_receive(struct mdns_socket *mdns_sock,
					   const char *name,
					   int query_type, int read_mode)
{
    /* wait up to ~1s */
    const useconds_t DELTA_T = 251000;
    const int RETRIES = 8;
    int retry = RETRIES;
    DNS_RECORD *rr_list;

    while (retry) {
	if (mdns_send_query(mdns_sock, name, query_type) == MDNS_STATUS_OK) {
	    rr_list = mdns_read_responses(mdns_sock->socket, read_mode, name);
	    if (rr_list != NULL) {
		DBG("send_and_receive: got response after %d retries\n",
		    RETRIES - retry);
		return rr_list;
	    }
	}
	/* MDNS servers delay responses. If the server just responded 
	 * to some query (maybe an earlier one we sent), we may need to wait. */
	--retry;
	usleep(DELTA_T);
    }
    BUG("send_and_receive: no response after %d retries\n", RETRIES);
    return NULL;
}

int mdns_probe_nw_scanners(char* uris_buf, int buf_size, int *count)
{
    int n = 0, bytes_read = 0;
    struct mdns_socket mdns_sock = MDNS_SOCKET_INIT;
    int stat = MDNS_STATUS_ERROR;
    DNS_RECORD *rr_list = NULL;
    const char scanner_name[] = "_scanner._tcp.local";

    DBG("mdns_probe_nw_scanners entry.\n");
    /* Open UDP socket */
    if (mdns_open_socket(&mdns_sock) != MDNS_STATUS_OK)
        goto bugout;

    rr_list = send_and_receive(&mdns_sock, scanner_name, QTYPE_PTR, MODE_READ_ALL);

    /* Update URIs buffer */
    bytes_read = mdns_update_uris(rr_list, uris_buf, buf_size, count);
    DBG("mdns_probe_nw_scanners returned with bytes_read = [%d].\n",bytes_read);

bugout:
    mdns_close_socket(&mdns_sock);

    mdns_rr_cleanup(rr_list);

    return bytes_read;
}

/*
 * Lookup IP for MDNS host name.
 * MDNS host name example: "npi7c8a3e" (LaserJet p2055dn)
 */
int mdns_lookup(char* hostname, unsigned char* ip)
{
    struct mdns_socket mdns_sock = MDNS_SOCKET_INIT;
    int stat = MDNS_STATUS_ERROR;
    char fqdn[MAX_NAME_LENGTH] = {0};
    DNS_RECORD *rr_list = NULL;

    DBG("mdns_probe_nw_scanners entry.\n");
    /* Open UDP socket */
    if (mdns_open_socket(&mdns_sock) != MDNS_STATUS_OK)
        goto bugout;

    /* Send dns query */
    sprintf(fqdn, "%s.local", hostname);

    rr_list = send_and_receive(&mdns_sock, fqdn, QTYPE_A, MODE_READ_SINGLE);

    /* Update IP Address buffer */
    if(rr_list)
    {
        strcpy(ip, rr_list->ip);
        stat = MDNS_STATUS_OK;
        DBG("IP = [%s].\n",ip);
    }

bugout:
    mdns_close_socket(&mdns_sock);

    mdns_rr_cleanup(rr_list);
    return stat;
}

